using ClientServer.Models.Users;

namespace ClientServer.Helpers
{
    /// <summary>
    /// these are the default system settings for <see cref="SystemSetting"/> (used as reset values)
    /// all settings here can be configured only through the frontend ui
    /// so these settings can be changed when the app is already running
    ///
    /// we use this values if the frontend requests resetting a value
    /// </summary>
    public static class InitialSystemSettings
    {
        /// <summary>
        /// the max amount of custom projects a user can create (this is inclusive!)
        /// IF THIS IS CHANGED AND a user has more projects he cannot create new custom projects (old projects are not touched)
        /// OR less than 0 for unlimited 
        /// </summary>
        public static int MaxCustomProjectsPerUser = 5;

        /// <summary>
        /// the max amount of custom tests for a participation for a user (this is inclusive)
        /// ALSO for custom projects!!
        /// IF THIS IS CHANGED AND a user has more custom tests he cannot create new tests (old tests are not touched)
        /// OR less than 0 for unlimited 
        /// </summary>
        public static int MaxCustomTestsPerParticipation = 7;

        /// <summary>
        /// a message that is displayed to all users (refreshed ever X seconds, X set in frontend)
        /// </summary>
        public static string CurrentServerMessage = "";

        /// <summary>
        /// the test server url, used to run tests
        /// </summary>
        public static string TestServerUrl = "http://localhost:8003/api.php";

        /// <summary>
        /// the test server stats url
        /// </summary>
        public static string TestServerStatsUrl = "http://localhost:8003/stats/";
        
        /// <summary>
        /// the submit test server stats url
        /// </summary>
        public static string SubmitTestServerStatsUrl = "http://localhost:8003/stats/";

        /// <summary>
        /// the submit test server url, used to run submit tests
        /// </summary>
        public static string SubmitTestServerUrl = "http://localhost:8003/api.php";

        /// <summary>
        /// the url to the ui to configure the test server
        /// </summary>
        public static string TestServerConfigUiUrl = "http://localhost:8000";
        
        /// <summary>
        /// the url to the ui to configure the submit test server
        /// </summary>
        public static string SubmitTestServerConfigUiUrl = "http://localhost:8000";

        /// <summary>
        ///  the timeout in ms to wait for a response from the test server before returning a testserver connection error
        /// </summary>
        public static int TestServerTimeoutInMs = 7000;

        /// <summary>
        ///  the timeout in ms to wait for a response from the test server before returning a testserver connection error
        /// </summary>
        public static int SubmitTestServerTimeoutInMs = 15000;


        /// <summary>
        /// the number of tests that can be run with one request (only the first X tests will be executed ... the other will be ignored
        /// <see cref="ResponseCode.PartialOk"/> will be returned in this case)
        /// </summary>
        public static int MaxNumberOfTestsWithOneRequest = 5;

        /// <summary>
        /// the number of tests that can be run with one request on the submit test server
        /// everything that uses the submit test server should make packages/buckets with request so that all tests will be run but with more than 1 request
        /// <see cref="ResponseCode.PartialOk"/> will be returned if there were more requested
        /// </summary>
        public static int MaxNumberOfTestsWithOneRequestSubmitTestServer = 15;

        
        /// <summary>
        /// the timeout for the just run program test runs
        /// </summary>
        public static int JustRunProgramTimeoutInMs = 3000;

        /// <summary>
        /// the timeout for the just run program test compiling
        /// </summary>
        public static int JustRunProgramCompileTimeoutInMs = 2000;
        
        /// <summary>
        /// the memory limit in kb for the just run program test runs
        /// </summary>
        public static int JustRunProgramMemoryLimitInKb = 2000;

        /// <summary>
        /// the max disk space for the just run program test runs
        /// </summary>
        public static int JustRunProgramMaxDiskSpaceInKb = 1000;

        
        /// <summary>
        /// the timeout for the custom project test runs
        /// </summary>
        public static int CustomProjectTestTimeoutInMs = 3000;

        /// <summary>
        /// the timeout for the custom project test compiling
        /// </summary>
        public static int CustomProjectTestCompileTimeoutInMs = 2000;
        
        /// <summary>
        /// the memory limit in kb for the custom project test runs
        /// </summary>
        public static int CustomProjectTestMemoryLimitInKb = 2000;

        /// <summary>
        /// the max disk space for the custom project test runs
        /// </summary>
        public static int CustomProjectTestMaxDiskSpaceInKb = 1000;
        
        /// <summary>
        /// <see cref="SystemSetting.MaxTestProtocolLinesToReturnInitially"/>
        /// </summary>
        public static int MaxTestProtocolLinesToReturnInitially = 400;

        /// <summary>
        /// the initial test server trace log api url
        /// </summary>
        public static string TestServerTraceLogApiUrl = "http://localhost:8003/worker/traceLog.php";
        


        //--- these cannot be used because the default role could have changed and deleted so this is would be invalid...
        
        /// <summary>
        /// the default group role for a user when added to a group / or when the group role was deleted then this default role is assigned
        /// </summary>
        public static int _DefaultGroupRoleId = 1;

        /// <summary>
        /// the default group for new users
        /// </summary>
        public static int _DefaultUserGroupId = 1;

        /// <summary>
        /// the default group when a user creates a new group
        /// </summary>
        public static int _DefaultGroupCreatorGroupRoleId = 2;
    }
}
