﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Http;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.Exercises
{
    /// <summary>
    /// controller for the known test types
    /// </summary>
    [Route(Constants.ApiPrefix + "langs")]
    public class LangController : ControllerWithDb
    {
        // static SemaphoreSlim semaphoreSlim = new SemaphoreSlim(1);

        public LangController(YapexDbContext context) : base(context)
        {
        }

        private async Task<bool> _CheckPermission()
        {
            if (!await base.IsLoggedIn()) return false;

            //user needs permission

            if (await HasSystemPermission(role => role != null && role.SystemRolePermission.CanChangeSystemSettings) ==
                false)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return false;
            }

            return true;
        }

        /// <summary>
        /// returns all known langs
        /// </summary>
        [HttpGet()]
        public async Task GetAll()
        {
            if (!await base.IsLoggedIn()) return;

            var langs = await _context.Langs.ToListAsync();

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponseWithData<List<LangFullBase>>
                (ResponseCode.Ok, "", langs.Select(p => new LangFullBase()
                {
                    Id = p.Id,
                    LangShortcut = p.LangShortcut,
                    DisplayName = p.Language
                }).ToList())));
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="lang"></param>
        /// <returns>the new lang</returns>
        [HttpPost("create")]
        public async Task CreateLang([FromBody] LangFullBase lang)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await this._CheckPermission()) return;

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "not supported yet")));

//            var newLang = new Lang()
//            {
//                Language = lang.DisplayName,
//                LangShortcut = lang.LangShortcut
//            };
//
//            try
//            {
//                _context.Langs.Add(newLang);
//                await _context.SaveChangesAsync();
//            }
//            catch (Exception e)
//            {
//                await base.HandleDbError(e);
//                return;
//            }
//
//
//            lang.Id = newLang.Id;
//            
//            await
//                Response.WriteAsync(Jc.Serialize(new BasicResponseWithData<LangFullBase>(ResponseCode.Ok,
//                    "", lang)));
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="lang"></param>
        /// <returns>the updated lang</returns>
        [HttpPost("update")]
        public async Task ChangeLang([FromBody] LangFullBase lang)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await this._CheckPermission()) return;

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "not supported yet")));

//            var oldLang = await _context.Langs.FirstOrDefaultAsync(p => p.Id == lang.Id);
//
//
//            if (oldLang == null)
//            {
//                await
//                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "language not found")));
//                return;
//            }
//
//
//            oldLang.Language = lang.DisplayName;
//            oldLang.LangShortcut = lang.LangShortcut;
//
//            try
//            {
//                await _context.SaveChangesAsync();
//            }
//            catch (Exception e)
//            {
//                await base.HandleDbError(e);
//                return;
//            }
//            
//            await
//                Response.WriteAsync(Jc.Serialize(new BasicResponseWithData<LangFullBase>(ResponseCode.Ok,
//                    "", lang)));
        }

        [HttpDelete("delete/{langId}")]
        public async Task DeleteLang(int langId)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await this._CheckPermission()) return;


            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "not supported yet")));

//            var oldLang = await _context.Langs
//                .FirstOrDefaultAsync(p => p.Id == langId);
//
//            if (oldLang == null)
//            {
//                await
//                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "language not found")));
//                return;
//            }
//
//
//            await semaphoreSlim.WaitAsync();
//
//            try
//            {
//                //we need at least one language...
//                var canDelete = false;
//
//                var langCount = _context.Langs.Count();
//
//                if (langCount > 1)
//                {
//                    canDelete = true;
//                }
//
//                if (canDelete == false)
//                {
//                    await
//                        Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
//                            "at least one language is needed")));
//                    return;
//                }
//
//                _context.Langs.Remove(oldLang);
//
//                await _context.SaveChangesAsync();
//            }
//            catch (Exception e)
//            {
//                await base.HandleDbError(e);
//                return;
//            }
//            finally
//            {
//                semaphoreSlim.Release();
//            }
//
//
//            await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }
    }


    public class LangFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// e.g. en/de...
        /// </summary>
        public string LangShortcut { get; set; }
    }
}
