﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;
using ClientServer.Helpers;
using ClientServer.Models;
using ClientServer.Models.CustomProjects;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.AfterSolutions;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Models.Files;
using ClientServer.Models.Users;
using ClientServer.Models.Users.Settings;
using Microsoft.EntityFrameworkCore;
using Microsoft.EntityFrameworkCore.Metadata;

namespace ClientServer.Db
{
    public class YapexDbContext : DbContext
    {

        /// <summary>
        /// the default max string length, some props might overwrite this setting
        /// then we should create a new var for them (so we don't lose overview)
        /// for normal text this should be enough
        ///
        /// this is not in constants because should not be changed...
        /// </summary>
        public const int DefaultMaxStringLength = 2000;

        public const int DescriptionContentMaxStringLength = 50000;

        public const int FileContentMaxStringLength = 80000;

        public const int TestContentMaxStringLength = 80000;
        
        public const int TestTimingsProtocolMaxStringLength = 80000;
        
        /// <summary>
        /// add 100 bytes for some header stuff e.g. exit code, timeouts, ...
        /// </summary>
        public const int TestProtocolMaxStringLength = TestContentMaxStringLength*2 + 100;

        /// <summary>
        /// also used for the server message
        /// </summary>
        public const int NotesMaxStringLength = 10000;
        
        
        /// <summary>
        /// property for every entity/table, the last datetime the entity was updated/changed
        /// 
        /// a shadow property 
        /// </summary>
        public static string LastUpdatedAt = "LastUpdatedAt";

        /// <summary>
        /// property for every entity/table, the datetime the entity was created
        /// 
        /// a shadow property 
        /// </summary>
        public static string CreatedAt = "CreatedAt";


        public DbSet<CustomProject> CustomProjects { get; set; }
        public DbSet<CustomProjectDescription> CustomProjectDescriptions { get; set; }
        public DbSet<CustomProjectSolution> CustomProjectSolutions { get; set; }
        public DbSet<CustomProjectSolutionFile> CustomProjectSolutionFiles { get; set; }
        public DbSet<CustomProjectTest> CustomProjectTests { get; set; }
        public DbSet<CustomProjectTestSettings> CustomProjectTestSettings { get; set; }
        public DbSet<CustomProjectTestWithSolutionAsTestResult> CustomProjectTestWithSolutionAsTestResults { get; set; }
        public DbSet<CustomProjectTestAsset> CustomProjectTestAssets { get; set; }
        
        public DbSet<ExerciseRelease> ExerciseReleases { get; set; }

        public DbSet<Solution> Solutions { get; set; }
        public DbSet<SolutionAssessment> SolutionAssessment { get; set; }
        public DbSet<SolutionFile> SolutionFiles { get; set; }
        
        public DbSet<TestWithSingleSolutionAsTestResult> TestWithSingleSolutionAsTestResult { get; set; }
        
        public DbSet<ExerciseReleaseWithUserAsParticipation> ExerciseReleaseWithUserAsParticipations { get; set; }
        
        public DbSet<CustomTestWithSingleSolutionAsTestResult> CustomTestWithSingleSolutionAsTestResult { get; set; }
        
        
        //---- files
        
        public DbSet<FileReferenceMarkdownAsset> FileReferenceMarkdownAssets { get; set; }
        public DbSet<ExerciseDescriptionWithFileAsAssetReference> ExerciseDescriptionWithFileAsAssetReferences { get; set; }
        
        public DbSet<FileReferenceTestAsset> FileReferenceTestAssets { get; set; }
        public DbSet<TestWithFileAsAssetReference> TestWithFileAsAssetReferences { get; set; }
        
        public DbSet<FileReferenceUserFileAsset> FileReferenceUserFileAssets { get; set; }
        public DbSet<CustomTestWithFileAsAssetReference> CustomTestWithFileAsAssetReferences { get; set; } //for custom tests files
        public DbSet<CustomProjectDescriptionWithFileAsAssetReference> CustomProjectDescriptionWithFileAsAssetReferences { get; set; } //custom project description files
        public DbSet<CustomProjectTestWithFileAsAssetReference> CustomProjectTestWithFileAsAssetReferences { get; set; } //custom project tests files
        
        //---- END files ----
        
        //--- after solution
        public DbSet<AfterSolution> AfterSolutions { get; set; }
        public DbSet<AfterSolutionFile> AfterSolutionFiles { get; set; }
        public DbSet<TestWithAfterSolutionAsTestResult> TestWithAfterSolutionAsTestResults { get; set; }
        public DbSet<CustomTestWithAfterSolutionAsTestResult> CustomTestWithAfterSolutionAsTestResults { get; set; }
        //---

        public DbSet<DefaultCustomTestSettings> DefaultCustomTestSettings { get; set; }
        //TODO MUST RENAME to CustomTests
        public DbSet<CustomTest> CustomTest { get; set; }
        public DbSet<Test> Tests { get; set; }
        public DbSet<TestSettings> TestCaseSettingses { get; set; }
        public DbSet<TestType> TestTypes { get; set; }


        public DbSet<Exercise> Exercises { get; set; }
        public DbSet<ExerciseDescription> ExerciseDescriptions { get; set; }
        public DbSet<MetaData> MetaDatas { get; set; }
        public DbSet<Tag> Tags { get; set; }
        public DbSet<TagWithMetaData> TagWithMetaDatas { get; set; }
        public DbSet<CodeTemplate> Templates { get; set; }
        //public DbSet<TopicGroup> TopicGroups { get; set; }

        public DbSet<GroupRole> GroupRoles { get; set; }
        public DbSet<GroupRolePermission> GroupRolePermissions { get; set; }

        public DbSet<OstepuExternalUser> ExternalUsers { get; set; }
        public DbSet<LDAPExternalUser> LDAPExternalUsers { get; set; }
        public DbSet<User> Users { get; set; }
        public DbSet<UserGroup> UserGroups { get; set; }


        public DbSet<UserSetting> UserSettingses { get; set; }

        public DbSet<UserWithUserGroup> UserWithUserGroups { get; set; }

        public DbSet<Lang> Langs { get; set; }
        public DbSet<PLang> PLangs { get; set; }

        public DbSet<CodeEditorSetting> CodeEditorSettings { get; set; }

        public DbSet<SystemRole> SystemRoles { get; set; }
        public DbSet<SystemRolePermission> SystemRolePermissions { get; set; }
        public DbSet<TemplateFile> TemplateFiles { get; set; }

        public DbSet<SystemSetting> SystemSettings { get; set; }
        public DbSet<AwaitDummy> AwaitDummies { get; set; }

        public DbSet<AuthToken> AuthTokens { get; set; }
        
        public static List<Type> AllTypes = new List<Type>()
        {
            typeof(CustomProject),
            typeof(CustomProjectDescription),
            typeof(CustomProjectSolution),
            typeof(CustomProjectSolutionFile),
            typeof(CustomProjectTest),
            typeof(CustomProjectTestSettings),
            typeof(CustomProjectTestWithSolutionAsTestResult),
            typeof(CustomProjectTestAsset),
            
            typeof(AfterSolution),
            typeof(AfterSolutionFile),
            typeof(TestWithAfterSolutionAsTestResult),
            typeof(CustomTestWithAfterSolutionAsTestResult),
            
            typeof(FileReferenceMarkdownAsset),
            typeof(FileReferenceTestAsset),
            typeof(FileReferenceUserFileAsset),
            
            typeof(ExerciseRelease),
            typeof(Solution),
            typeof(SolutionAssessment),
            typeof(SolutionFile),
            typeof(TestWithSingleSolutionAsTestResult),
            typeof(ExerciseReleaseWithUserAsParticipation),
            typeof(CustomTestWithSingleSolutionAsTestResult),
            typeof(DefaultCustomTestSettings),
            typeof(CustomTest),
            typeof(Test),
            typeof(TestSettings),
            typeof(TestType),
            typeof(Exercise),
            typeof(ExerciseDescription),
            typeof(MetaData),
            typeof(Tag),
            typeof(TagWithMetaData),
            typeof(CodeTemplate),
            typeof(GroupRole),
            typeof(GroupRolePermission),
            typeof(OstepuExternalUser),
            typeof(LDAPExternalUser),
            typeof(User),
            typeof(UserGroup),
            typeof(UserSetting),
            typeof(UserWithUserGroup),
            typeof(Lang),
            typeof(PLang),
            typeof(CodeEditorSetting),
            typeof(SystemRole),
            typeof(SystemRolePermission),
            typeof(TemplateFile),
            typeof(SystemSetting),
            typeof(AuthToken),
        };

        public YapexDbContext(DbContextOptions options) : base(options)
        {
        }


        protected override void OnModelCreating(ModelBuilder m)
        {
            m.Entity<AwaitDummy>();            

            #region custom project

            //no need to configure principal tables
            //dependent table config

            m.Entity<CustomProject>()
                .HasOne(p => p.User)
                .WithMany(p => p.CustomProjects)
                .HasForeignKey(p => p.UserId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            m.Entity<CustomProject>()
                .HasMany(p => p.Solutions)
                .WithOne(p => p.CustomProject)
                .HasForeignKey(p => p.CustomProjectId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<CustomProject>()
                .HasMany(p => p.Tests)
                .WithOne(p => p.CustomProject)
                .HasForeignKey(p => p.CustomProjectId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<CustomProjectSolution>()
                .HasMany(p => p.SolutionFiles)
                .WithOne()
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            m.Entity<CustomProjectSolution>()
                .HasOne(p => p.MainFile)
                .WithOne()
                .HasForeignKey<CustomProjectSolution>(p => p.MainFileId)
                .HasConstraintName("FK_CustomProjectSolutionFiles_CustomProjectSolutions_CustomProj")
                .IsRequired(false)
                .OnDelete(DeleteBehavior.SetNull)
                ;
            
            m.Entity<CustomProject>()
                .HasOne(p => p.Description)
                .WithOne(p => p.CustomProject)
                .HasForeignKey<CustomProjectDescription>(p => p.CustomProjectId)
                .HasConstraintName("FK_CustomProjectDescriptionWithFileAsAssetReferences_CustomProj")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #region custom project descirption asset referneces join table

            m.Entity<CustomProjectDescriptionWithFileAsAssetReference>()
                .HasKey(p => new {p.CustomProjectDescriptionId, p.FileReferenceUserFileAssetId})
                ;

            m.Entity<CustomProjectDescriptionWithFileAsAssetReference>()
                .HasIndex(p => p.CustomProjectDescriptionId)
                .HasName("IX_CustomProjectDescriptionWithFileAsAssetReferences_CustomProj")
                ;
            
            m.Entity<CustomProjectDescriptionWithFileAsAssetReference>()
                .HasIndex(p => p.FileReferenceUserFileAssetId)
                .HasName("IX_CustomProjectDescriptionWithFileAsAssetReferences_FileRefere")
                ;

            m.Entity<CustomProjectDescriptionWithFileAsAssetReference>()
                .HasOne(p => p.CustomProjectDescription)
                .WithMany(p => p.AssetReferences)
                .HasConstraintName("FK_CustomProjectDescriptionWithFileAsAssetReferences_CustomProj")
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            m.Entity<CustomProjectDescriptionWithFileAsAssetReference>()
                .HasOne(p => p.FileReferenceUserFileAsset)
                .WithMany(p => p.CustomProjectDescriptionWithFileAsAssetReferences)
                .HasConstraintName("FK_CustomProjectDescriptionWithFileAsAssetReferences_FileRefere")
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            m.Entity<CustomProjectTest>()
                .HasOne(p => p.CustomProjectTestSettings)
                .WithOne(p => p.CustomProjectTest)
                .HasForeignKey<CustomProjectTestSettings>(p => p.CustomProjectTestId)
                .HasConstraintName("FK_CustomProjectTestSettings_CustomProjectTests_CustomProjectTe")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<CustomProjectTestSettings>()
                .HasKey(p => p.Id);


            #region custom project test asset references join table


            m.Entity<CustomProjectTestWithFileAsAssetReference>()
                .HasKey(p => new {p.CustomProjectTestId, p.FileReferenceUserFileAssetId})
                ;

            m.Entity<CustomProjectTestWithFileAsAssetReference>()
                .HasIndex(p => p.CustomProjectTestId)
                .HasName("IX_CustomProjectTestWithFileAsAssetReferences_CustomProjectTest")
                ;

            m.Entity<CustomProjectTestWithFileAsAssetReference>()
                .HasIndex(p => p.FileReferenceUserFileAssetId)
                .HasName("IX_CustomProjectTestWithFileAsAssetReferences_FileReferenceUser")
                ;
            
            m.Entity<CustomProjectTestWithFileAsAssetReference>()
                .HasOne(p => p.CustomProjectTest)
                .WithMany(p => p.AssetReferences)
                .HasConstraintName("FK_CustomProjectTestWithFileAsAssetReferences_CustomProjectTest")
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<CustomProjectTestWithFileAsAssetReference>()
                .HasOne(p => p.FileReferenceUserFileAsset)
                .WithMany(p => p.CustomProjectTestWithFileAsAssetReferences)
                .HasConstraintName("FK_CustomProjectTestWithFileAsAssetReferences_FileReferenceUser")
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            #region  n:m tables

            m.Entity<CustomProjectTestWithSolutionAsTestResult>()
                .HasKey(p => new {p.CustomProjectSolutionId, p.CustomProjectTestId})
                ;
            
            m.Entity<CustomProjectTestWithSolutionAsTestResult>()
                .HasIndex(p => p.CustomProjectSolutionId)
                .HasName("IX_CustomProjectTestWithSolutionAsTestResults_CustomProjectSolu") //ef 3 would truncate the last char and replace it with ~ but ef core 1 would just write the last char this happens after >= 63 characters 
                ;
            
            m.Entity<CustomProjectTestWithSolutionAsTestResult>()
                .HasIndex(p => p.CustomProjectTestId)
                .HasName("IX_CustomProjectTestWithSolutionAsTestResults_CustomProjectTest")
                ;

            //no need to configure principal tables

            //n:n table CustomProjectTestWithSolutionAsTestResult
            m.Entity<CustomProjectTestWithSolutionAsTestResult>()
                .HasOne(p => p.CustomProjectSolution)
                .WithMany(p => p.TestResults)
                .HasForeignKey(p => p.CustomProjectSolutionId)
                .HasConstraintName("FK_CustomProjectTestWithSolutionAsTestResults_CustomProjectSolu")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            m.Entity<CustomProjectTestWithSolutionAsTestResult>()
                .HasOne(p => p.CustomProjectTest)
                .WithMany(p => p.TestResults)
                .HasForeignKey(p => p.CustomProjectTestId)
                .HasConstraintName("FK_CustomProjectTestWithSolutionAsTestResults_CustomProjectTest")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion
            
            #endregion
            
            #region exercise_focus_misc

            m.Entity<Exercise>()
                .HasMany(p => p.CodeTemplates)
                .WithOne(p => p.Exercise)
                .HasForeignKey(p => p.ExerciseId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<CodeTemplate>()
                .HasOne(p => p.PLang)
                .WithMany()
                .HasForeignKey(p => p.PLangId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Restrict)
                ;

            m.Entity<CodeTemplate>()
                .HasOne(p => p.MainFile)
                .WithOne()
                .HasForeignKey<CodeTemplate>(p => p.MainFileId)
                .IsRequired(false) //true would be better but ef cant handle save (circular dependency)
                .OnDelete(DeleteBehavior.SetNull) //TODO MUST??
                ;

            m.Entity<CodeTemplate>()
                .HasMany(p => p.TemplateFiles)
                .WithOne()
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<Exercise>()
                .HasOne(p => p.Description)
                .WithOne(p => p.Exercise)
                .HasForeignKey<ExerciseDescription>(p => p.ExerciseId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #region exercise description asset references join table
            
            //see https://www.learnentityframeworkcore.com/configuration/many-to-many-relationship-configuration
            /*
             modelBuilder.Entity<BookCategory>()
                .HasKey(bc => new { bc.BookId, bc.CategoryId });  
        
            modelBuilder.Entity<BookCategory>()
                .HasOne(bc => bc.Book)
                .WithMany(b => b.BookCategories)
                .HasForeignKey(bc => bc.BookId);  
            
            modelBuilder.Entity<BookCategory>()
                .HasOne(bc => bc.Category)
                .WithMany(c => c.BookCategories)
                .HasForeignKey(bc => bc.CategoryId);
                 */

            m.Entity<ExerciseDescriptionWithFileAsAssetReference>()
                .HasKey(p => new {p.ExerciseDescriptionId, p.FileReferenceMarkdownAssetId})
                ;

            m.Entity<ExerciseDescriptionWithFileAsAssetReference>()
                .HasIndex(p => p.ExerciseDescriptionId)
                .HasName("IX_ExerciseDescriptionWithFileAsAssetReferences_ExerciseDescrip")
                ;
            
            m.Entity<ExerciseDescriptionWithFileAsAssetReference>()
                .HasIndex(p => p.FileReferenceMarkdownAssetId)
                .HasName("IX_ExerciseDescriptionWithFileAsAssetReferences_FileReferenceMa")
                ;

            m.Entity<ExerciseDescriptionWithFileAsAssetReference>()
                .HasOne(p => p.ExerciseDescription)
                .WithMany(p => p.AssetReferences)
                .HasConstraintName("FK_ExerciseDescriptionWithFileAsAssetReferences_ExerciseDescrip")
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            m.Entity<ExerciseDescriptionWithFileAsAssetReference>()
                .HasOne(p => p.FileReferenceMarkdownAsset)
                .WithMany(p => p.AssetReferences)
                .HasConstraintName("FK_ExerciseDescriptionWithFileAsAssetReferences_FileReferenceMa")
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            //old fk ExerciseId1 --> delete...
            m.Entity<Exercise>()
                .HasOne(p => p.MetaData)
                .WithOne(p => p.Exercise)
                .HasForeignKey<MetaData>(p => p.ExerciseId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            //hasindex ensures unique value BUT can be modified (alternate keys not)
            //see https://github.com/aspnet/EntityFramework/issues/4914
            //this is not really working ... check in controllers..
            m.Entity<Tag>()
                .HasIndex(p => p.DisplayName)
                .HasName("Unique_Tag")
                ;

            #region config tags n:n tables

            m.Entity<TagWithMetaData>()
                .HasKey(p => new {p.TagId, p.MetaDataId})
                ;

            m.Entity<TagWithMetaData>()
                .HasIndex(p => p.MetaDataId)
                ;
                
            m.Entity<TagWithMetaData>()
                .HasIndex(p => p.TagId)
                ;

            //no need to configure principal tables
            //m.Entity<MetaData>()
            //    .HasMany(p => p.TagWithMetaDatas)
            //    .WithOne(p => p.MetaData)
            //    .HasForeignKey(p => new {p.TagId, p.MetaDataId})
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //m.Entity<Tag>()
            //    .HasMany(p => p.TagWithMetaDatas)
            //    .WithOne(p => p.Tag)
            //    .HasForeignKey(p => new {p.TagId, p.MetaDataId})
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //n:n table for tags/meta data
            m.Entity<TagWithMetaData>()
                .HasOne(p => p.MetaData)
                .WithMany(p => p.TagWithMetaDatas)
                .HasForeignKey(p => p.MetaDataId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            //TODO delete if working but keep docs
//why? we may want to use that tags also for custom projects... & we don't need this!
//but this can be done for a N:M join table
//            m.Entity<TagWithMetaData>()
//                .HasOne(p => p.Tag)
//                .WithMany(p => p.TagWithMetaDatas)
//                .HasForeignKey(p => p.TagId)
//                .IsRequired()
//                .OnDelete(DeleteBehavior.Cascade)
//                ;

            #endregion

            //MAYBE 
            //lang and content type is encoded in the content string
            /*
            m.Entity<ExerciseDescription>()
                .HasOne(p => p.Lang)
                .WithOne()
                .IsRequired()
                .HasForeignKey<ExerciseDescription>(p => p.LangId)
                .OnDelete(DeleteBehavior.SetNull)
                ;

            m.Entity<ExerciseDescription>()
                .HasOne(p => p.ContentType)
                .WithOne()
                .IsRequired()
                .HasForeignKey<ExerciseDescription>(p => p.ContentTypeId)
                .OnDelete(DeleteBehavior.SetNull)
                ;

            */

            #endregion

            #region exercise_focus_tests

            #region test

            m.Entity<Exercise>()
                .HasMany(p => p.Tests)
                .WithOne(p => p.Exercise)
                .HasForeignKey(p => p.ExerciseId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            #region exercise test files join table
            
            //see https://www.learnentityframeworkcore.com/configuration/many-to-many-relationship-configuration

            #region exercise test asset references join table

            m.Entity<TestWithFileAsAssetReference>()
                .HasKey(p => new {p.TestId, p.FileReferenceTestAssetId})
                ;

            m.Entity<TestWithFileAsAssetReference>()
                .HasIndex(p => p.FileReferenceTestAssetId)
                ;
            
            m.Entity<TestWithFileAsAssetReference>()
                .HasIndex(p => p.TestId)
                ;

            m.Entity<TestWithFileAsAssetReference>()
                .HasOne(p => p.Test)
                .WithMany(p => p.AssetReferences)
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            m.Entity<TestWithFileAsAssetReference>()
                .HasOne(p => p.FileReferenceTestAsset)
                .WithMany(p => p.AssetReferences)
                .HasConstraintName("FK_TestWithFileAsAssetReferences_FileReferenceTestAssets_FileRe")
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion
          
            
            
            #endregion

            m.Entity<Test>()
                .HasOne(p => p.TestSettings)
                .WithOne(p => p.Test)
                .HasForeignKey<TestSettings>(p => p.TestId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<Test>()
                .HasOne(p => p.TestType)
                .WithMany()
                .HasForeignKey(p => p.TestTypeId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Restrict) //TODO not applied??
                ;

            #endregion

            m.Entity<Exercise>()
                .HasOne(p => p.DefaultCustomTestSettings)
                .WithOne(p => p.Exercise)
                .HasForeignKey<DefaultCustomTestSettings>(p => p.ExerciseId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #region exercise release

            m.Entity<ExerciseRelease>()
               .HasAlternateKey(p => p.GeneratedCode)
               .HasName("Unique_GeneratedCode")
               ;

            m.Entity<Exercise>()
                .HasMany(p => p.Releases)
                .WithOne(p => p.Exercise)
                .HasForeignKey(p => p.ExerciseId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Restrict)
                ;


            m.Entity<ExerciseRelease>()
                .HasOne(p => p.PLang)
                .WithMany()
                .HasForeignKey(p => p.PLangId)
                .IsRequired() 
                .OnDelete(DeleteBehavior.Restrict)
                ;

            #endregion

            #region release - participation - user

            //composite primary key
            //n:n table for exerciseRelease & user (store the exercise soltuion for the release)
            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasKey(p => new { p.UserId, p.ExerciseReleaseId})
                ;

            //no need to configure principal tables
            //m.Entity<ExerciseRelease>()
            //    .HasMany(p => p.ExerciseReleaseWithUserAsParticipations)
            //    .WithOne(p => p.ExerciseRelease)
            //    .IsRequired()
            //    .HasForeignKey(p => new {p.ExerciseReleaseId, p.UserId})
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;


            //m.Entity<User>()
            //    .HasMany(p => p.ExerciseReleaseWithUserAsParticipations)
            //    .WithOne(p => p.User)
            //    .HasForeignKey(p => new {p.ExerciseReleaseId, p.UserId})
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //dependent table config
            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasOne(p => p.ExerciseRelease)
                .WithMany(p => p.ExerciseReleaseWithUserAsParticipations)
                .HasForeignKey(p => p.ExerciseReleaseId)
                .HasConstraintName("FK_ExerciseReleaseWithUserAsParticipations_ExerciseReleases_Exe")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            //dependent table config
            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasOne(p => p.User)
                .WithMany(p => p.ExerciseReleaseWithUserAsParticipations)
                .HasForeignKey(p => p.UserId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            #region participation / solution

            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasOne(p => p.LastEditedPLang)
                .WithMany()
                .HasForeignKey(p => p.LastEditedPLangId)
                .HasConstraintName("FK_ExerciseReleaseWithUserAsParticipations_PLangs_LastEditedPLa")
                .IsRequired()
                .OnDelete(DeleteBehavior.Restrict)
                ;

            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasMany(p => p.Solutions)
                .WithOne(p => p.ExerciseReleaseWithUserAsParticipation)
                .HasForeignKey(p => new {p.UserId, p.ExerciseReleaseId}) //order matters!! must match ExerciseReleaseWithUserAsParticipation primary key (pk)
                .HasConstraintName("FK_Solutions_ExerciseReleaseWithUserAsParticipations_UserId_Exe")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<Solution>()
                .HasKey(p => new {p.UserId, p.ExerciseReleaseId, p.PLangId})
                ;

            m.Entity<Solution>()
                .HasIndex(p => new {p.UserId, p.ExerciseReleaseId})
                ;


            m.Entity<TemplateFile>()
                .HasMany(p => p.SolutionParts)
                .WithOne(p => p.TemplateFile)
                .HasForeignKey(p => p.TemplateFileId)
                .IsRequired(false)
                .OnDelete(DeleteBehavior.SetNull)
                ;

            m.Entity<Solution>()
                .HasOne(p => p.PLang)
                .WithMany()
                .HasForeignKey(p => p.PLangId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Restrict)
                ;

            #region NEW changed solution (now withour versions)
                
            m.Entity<Solution>()
                .HasMany(p => p.SolutionFiles)
                .WithOne(p => p.Solution)
                .HasForeignKey(p => new {p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                .HasConstraintName("FK_SolutionFiles_Solutions_SolutionUserId_SolutionExerciseRelea")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<Solution>()
                .HasOne(p => p.MainFile)
                .WithOne()
                .HasForeignKey<Solution>(p => p.MainFileId)
                .IsRequired(false)
                .OnDelete(DeleteBehavior.SetNull)
                ;

            
            #endregion

            //assessment
            m.Entity<SolutionAssessment>()
                .HasOne(p => p.Solution)
                .WithOne(p => p.Assessment)
                .IsRequired()
                .OnDelete(DeleteBehavior.SetNull)
                ;

            m.Entity<Solution>()
                .HasOne(p => p.Assessment)
                .WithOne(p => p.Solution)
                .HasForeignKey<SolutionAssessment>(p => new {p.UserId, p.ExerciseReleaseId, p.PLangId}) //added
                .HasConstraintName("FK_SolutionAssessment_Solutions_UserId_ExerciseReleaseId_PLangI")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            m.Entity<SolutionFile>()
                .HasKey(p => p.Id)
                ;
            
            m.Entity<SolutionFile>()
                .HasIndex(p => new {p.TemplateFileId})
                ;
            
            m.Entity<SolutionFile>()
                .HasIndex(p => new {p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                .HasName("IX_SolutionFiles_SolutionUserId_SolutionExerciseReleaseId_Solut")
                ;

            #region after solution

            m.Entity<Solution>()
                .HasOne(p => p.AfterSolution)
                .WithOne(p => p.Solution)
                .HasForeignKey<AfterSolution>(p => new {p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                .HasConstraintName("FK_AfterSolutions_Solutions_SolutionUserId_SolutionExerciseRele")
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<AfterSolution>()
                .HasMany(p => p.SolutionFiles)
                .WithOne(p => p.AfterSolution)
                .HasForeignKey(p => p.AfterSolutionId)
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<AfterSolution>()
                .HasOne(p => p.MainFile)
                .WithOne()
                .HasForeignKey<AfterSolution>(p => p.MainFileId)
                .IsRequired(false)
                .OnDelete(DeleteBehavior.SetNull)
                ;

            //--- normal tests

            m.Entity<TestWithAfterSolutionAsTestResult>()
                .HasKey(p => new { p.TestId, p.AfterSolutionId})
                ;

            m.Entity<TestWithAfterSolutionAsTestResult>()
                .HasIndex(p => p.AfterSolutionId)
                ;
            
            m.Entity<TestWithAfterSolutionAsTestResult>()
                .HasIndex(p => p.TestId)
                ;
            
            m.Entity<TestWithAfterSolutionAsTestResult>()
                .HasOne(p => p.Test)
                .WithMany(p => p.AfterTestResults)
                .HasForeignKey(p => p.TestId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<TestWithAfterSolutionAsTestResult>()
                .HasOne(p => p.AfterSolution)
                .WithMany(p => p.TestResults)
                .HasForeignKey(p => p.AfterSolutionId)
                .HasConstraintName("FK_TestWithAfterSolutionAsTestResults_AfterSolutions_AfterSolut")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;
            
            //custom tests

            m.Entity<CustomTestWithAfterSolutionAsTestResult>()
                .HasKey(p => new {p.CustomTestId, p.AfterSolutionId })
                ;
            
            m.Entity<CustomTestWithAfterSolutionAsTestResult>()
                .HasIndex(p => p.AfterSolutionId)
                ;
            
            m.Entity<CustomTestWithAfterSolutionAsTestResult>()
                .HasIndex(p => p.CustomTestId)
                ;

            m.Entity<CustomTestWithAfterSolutionAsTestResult>()
                .HasOne(p => p.CustomTest)
                .WithMany(p => p.AfterTestResults)
                .HasForeignKey(p => p.CustomTestId)
                .HasConstraintName("FK_CustomTestWithAfterSolutionAsTestResults_CustomTest_CustomTe")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<CustomTestWithAfterSolutionAsTestResult>()
                .HasOne(p => p.AfterSolution)
                .WithMany(p => p.CustomTestResults)
                .HasForeignKey(p => p.AfterSolutionId)
                .HasConstraintName("FK_CustomTestWithAfterSolutionAsTestResults_AfterSolutions_Afte")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            #region (all kind) test results all n:n tables

            //.HasForeignKey(p => new {p.UserId, p.ExerciseReleaseId, p.PLangId})
            // order matters here!!!! must be the same as primary key from solution!!

            #region normal/submit tests

           
            m.Entity<TestWithSingleSolutionAsTestResult>()
                .HasKey(p => new {p.TestId, p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                ;

            m.Entity<TestWithSingleSolutionAsTestResult>()
                .HasIndex(p => p.TestId)
                ;
            
            m.Entity<TestWithSingleSolutionAsTestResult>()
                .HasIndex(p => new {p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                .HasName("IX_TestWithSingleSolutionAsTestResult_SolutionUserId_SolutionEx")
                ;
         
            //no need to configure principal tables
            //m.Entity<Test>()
            //    .HasMany(p => p.TestResults)
            //    .WithOne(p => p.Test)
            //    .HasForeignKey(p => new { p.TestId, p.SolutionVersionId })
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //m.Entity<SolutionVersion>()
            //    .HasMany(p => p.TestResults)
            //    .WithOne(p => p.SolutionVersion)
            //    .HasForeignKey(p => new { p.TestId, p.SolutionVersionId })
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //dependent n:n table
            m.Entity<TestWithSingleSolutionAsTestResult>()
                .HasOne(p => p.Test)
                .WithMany(p => p.TestResultsNew)
                .HasForeignKey(p => p.TestId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;


            //dependent n:n table
            m.Entity<TestWithSingleSolutionAsTestResult>()
                .HasOne(p => p.Solution)
                .WithMany(p => p.TestResults)
                .HasForeignKey(p => new {p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                .HasConstraintName("FK_TestWithSingleSolutionAsTestResult_Solutions_SolutionUserId_")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;


            #endregion


            #region custom tests

            m.Entity<CustomTestWithSingleSolutionAsTestResult>()
                .HasKey(p => new { p.CustomTestId, p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                ;

            m.Entity<CustomTestWithSingleSolutionAsTestResult>()
                .HasIndex(p => p.CustomTestId)
                ;
            
            m.Entity<CustomTestWithSingleSolutionAsTestResult>()
                .HasIndex(p => new {p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                .HasName("IX_CustomTestWithSingleSolutionAsTestResult_SolutionUserId_Solu")
                ;

            //no need to configure principal tables
            //m.Entity<CustomTest>()
            //    .HasMany(p => p.TestResults)
            //    .WithOne(p => p.CustomTest)
            //    .HasForeignKey(p => new { p.CustomTestId, p.SolutionVersionId })
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //m.Entity<SolutionVersion>()
            //    .HasMany(p => p.CustomTestResults)
            //    .WithOne(p => p.SolutionVersion)
            //    .HasForeignKey(p => new { p.CustomTestId, p.SolutionVersionId })
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //n:n table
            m.Entity<CustomTestWithSingleSolutionAsTestResult>()
                .HasOne(p => p.CustomTest)
                .WithMany(p => p.TestResultsNew)
                .HasForeignKey(p => p.CustomTestId)
                .HasConstraintName("FK_CustomTestWithSingleSolutionAsTestResult_CustomTest_CustomTe")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;


            //n:n
            m.Entity<CustomTestWithSingleSolutionAsTestResult>()
                .HasOne(p => p.Solution)
                .WithMany(p => p.CustomTestResults)
                .HasForeignKey(p => new {p.SolutionUserId, p.SolutionExerciseReleaseId, p.SolutionPLangId})
                .HasConstraintName("FK_CustomTestWithSingleSolutionAsTestResult_Solutions_SolutionU")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            #endregion


            #region participation / custom tests

            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasMany(p => p.CustomTests)
                .WithOne() //pk of ExerciseReleaseWithUserAsParticipation
                .HasForeignKey(p => new { p.UserId, p.ExerciseReleaseId })
                .HasConstraintName("FK_CustomTest_ExerciseReleaseWithUserAsParticipations_UserId_Ex")
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasIndex(p => p.ExerciseReleaseId)
                ;
            
            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasIndex(p => p.LastEditedPLangId)
                ;
            
            m.Entity<ExerciseReleaseWithUserAsParticipation>()
                .HasIndex(p => p.UserId)
                ;

            m.Entity<CustomTest>()
                .HasOne(p => p.TestType)
                .WithMany()
                .HasForeignKey(p => p.TestTypeId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;
           

            #region custom test asset references

            m.Entity<CustomTestWithFileAsAssetReference>()
                .HasKey(p => new {p.CustomTestId, p.FileReferenceUserFileAssetId})
                ;

            m.Entity<CustomTestWithFileAsAssetReference>()
                .HasIndex(p => p.CustomTestId)
                ;
            
            m.Entity<CustomTestWithFileAsAssetReference>()
                .HasIndex(p => p.FileReferenceUserFileAssetId)
                .HasName("IX_CustomTestWithFileAsAssetReferences_FileReferenceUserFileAss")
                ;

            m.Entity<CustomTestWithFileAsAssetReference>()
                .HasOne(p => p.CustomTest)
                .WithMany(p => p.AssetReferences)
                .OnDelete(DeleteBehavior.Cascade)
                ;
                
            m.Entity<CustomTestWithFileAsAssetReference>()
                .HasOne(p => p.FileReferenceUserFileAsset)
                .WithMany(p => p.CustomTestWithFileAsAssetReferences)
                .HasConstraintName("FK_CustomTestWithFileAsAssetReferences_FileReferenceUserFileAss")
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            #endregion

            #endregion


            #region users and permission


            //alternate keys cannot be changed... 
            //see https://github.com/aspnet/EntityFramework/issues/4073

            m.Entity<User>()
               .HasAlternateKey(p => p.Token)
               .HasName("Unique_Token")
               ;

            m.Entity<User>()
              .HasMany(p => p.Exercises)
              .WithOne(p => p.User)
              .HasForeignKey(p => p.UserId)
              .IsRequired(false)
              .OnDelete(DeleteBehavior.SetNull) //do not delete exercises when the creator is deleted...
              ;

            m.Entity<User>()
                .HasOne(p => p.UserSettings)
                .WithOne(p => p.User)
                .HasForeignKey<UserSetting>(p => p.UserId) //if deleted (null) use some defaults
                .IsRequired(false)
                .OnDelete(DeleteBehavior.Cascade)
                ;


            #region user settings

            m.Entity<UserSetting>()
              .HasOne(p => p.CodeEditorSetting)
              .WithOne(p => p.UserSetting)
              .HasForeignKey<CodeEditorSetting>(p => p.UserSettingId)
              .IsRequired()
              .OnDelete(DeleteBehavior.Cascade)
              ;

            //DeleteBehavior.SetNull and .IsRequired() is not a good idea...
            //because IsRequired() forces not NULL and setNull tries to set to null...
            //so we cannot remove the principal... e.g. lang
            m.Entity<UserSetting>()
                .HasOne(p => p.Lang)
                .WithMany()
                .HasForeignKey(p => p.LangId)
                .IsRequired(false)
                .OnDelete(DeleteBehavior.SetNull)
                ;

            #endregion

            #region user / group 

            m.Entity<UserGroup>()
               .HasMany(p => p.Exercises)
               .WithOne(p => p.UserGroup)
               .HasForeignKey(p => p.UserGroupId)
               .IsRequired()
               .OnDelete(DeleteBehavior.Cascade)
               ;

            //composite primary key
            m.Entity<UserWithUserGroup>()
                .HasKey(p => new {p.UserId, p.UserGroupId})
                ;

            m.Entity<UserWithUserGroup>()
                .HasIndex(p => p.GroupRoleId)
                ;
            
            m.Entity<UserWithUserGroup>()
                .HasIndex(p => p.UserGroupId)
                ;
            
            m.Entity<UserWithUserGroup>()
                .HasIndex(p => p.UserId)
                ;

            //no need to configure principal tables
            //m.Entity<User>()
            //    .HasMany(p => p.UserWithUserGroups)
            //    .WithOne(p => p.User)
            //    .HasForeignKey(p => new {p.UserId})
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade)
            //    ;

            //m.Entity<UserGroup>()
            //    .HasMany(p => p.UserWithUserGroups)
            //    .WithOne(p => p.UserGroup)
            //    .HasForeignKey(p => new { p.UserGroupId })
            //    .IsRequired()
            //    .OnDelete(DeleteBehavior.Cascade) //delete userUserGroup when the userGroup is deleted?
            //    ;

            //n:n table
            m.Entity<UserWithUserGroup>()
                .HasOne(p => p.User)
                .WithMany(p => p.UserWithUserGroups)
                .HasForeignKey(p => p.UserId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            //n:n table
            m.Entity<UserWithUserGroup>()
                .HasOne(p => p.UserGroup)
                .WithMany(p => p.UserWithUserGroups)
                .HasForeignKey(p => p.UserGroupId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<User>()
                .HasOne(p => p.OstepuExternalUser)
                .WithOne(p => p.User)
                .HasForeignKey<OstepuExternalUser>(p => p.UserId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<OstepuExternalUser>()
                .HasKey(p => p.ExternalId)
                ;

            m.Entity<OstepuExternalUser>()
                .HasAlternateKey(p => p.Token)
                .HasName("Unique_ExternalToken")
                ;
            
            m.Entity<User>()
                .HasOne(p => p.LDAPExternalUser)
                .WithOne(p => p.User)
                .HasForeignKey<LDAPExternalUser>(p => p.UserId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            m.Entity<LDAPExternalUser>()
                .HasKey(p => p.ExternalId)
                ;

            #endregion

            #region group / group role

            m.Entity<GroupRole>()
               .HasMany(p => p.UserUserGroups)
               .WithOne(p => p.GroupRole)
//               .HasForeignKey(p => p.GroupRoleId)
               .IsRequired()
                //Cascade or restrict?? cascade = user is kicked out of group if role is deleted
                //restrict: role can only be deleted if no user has this role
               .OnDelete(DeleteBehavior.Restrict)
               ;

            m.Entity<GroupRole>()
                .HasOne(p => p.GroupRolePermission)
                .WithOne(p => p.GroupRole)
                .HasForeignKey<GroupRolePermission>(p => p.GroupRoleId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            #region user / system role

            m.Entity<User>()
                .HasOne(p => p.SystemRole)
                .WithMany()
                .HasForeignKey(p => p.SystemRoleId)
                .IsRequired(false)
                .OnDelete(DeleteBehavior.SetNull) //when we delete a system role we downgrade all users with this role
                ;

            m.Entity<SystemRole>()
                .HasOne(p => p.SystemRolePermission)
                .WithOne(p => p.SystemRole)
                .HasForeignKey<SystemRolePermission>(p => p.SystemRoleId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade)
                ;

            #endregion

            #region system settings (e.g. default group role, default user group ...)

            m.Entity<SystemSetting>()
                .HasOne(p => p.DefaultUserGroup)
                .WithOne()
                .HasForeignKey<SystemSetting>(p => p.DefaultUserGroupId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Restrict)
                ;

            m.Entity<SystemSetting>()
                .HasOne(p => p.DefaultGroupRole)
                .WithOne()
                .HasForeignKey<SystemSetting>(p => p.DefaultGroupRoleId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Restrict)
                ;

            m.Entity<SystemSetting>()
                .HasOne(p => p.DefaultGroupCreatorGroupRole)
                .WithOne()
                .IsRequired()
                .HasForeignKey<SystemSetting>(p => p.DefaultGroupCreatorGroupRoleId)
                .OnDelete(DeleteBehavior.Restrict)
                ;

            #endregion


            #endregion


            m.Entity<AuthToken>()
                .HasKey(p => p.UserId)
                ;

            m.Entity<AuthToken>()
                .HasIndex(p => p.UserId)
                .IsUnique()
                ;

            m.Entity<AuthToken>()
                .HasIndex(p => p.UserAuthToken)
                .IsUnique()
                ;

            m.Entity<AuthToken>()
                .HasOne(p => p.User)
                .WithOne()
                .HasForeignKey<AuthToken>(p => p.UserId)
                .IsRequired()
                .OnDelete(DeleteBehavior.Cascade) //TOOD is the user really th principal?? else cascade would be bad!
                ;
               

           var entityMethod =
                m.GetType()
                    .GetMethods()
                    .First(p => p.Name == "Entity" && p.IsGenericMethod && p.GetParameters().Length == 0);

            //reflection is ok here because this is only called one time
            //add timestamps to all entities
            foreach (var type in AllTypes)
            {
                var method = entityMethod.MakeGenericMethod(type);
                var result = method.Invoke(m, new object[] {});

                var propertyMethod = result.GetType()
                    .GetMethods()
                    .First(
                        p =>
                            p.Name == "Property" && p.IsGenericMethod && p.GetParameters().Length == 1 &&
                            p.GetParameters()[0].ParameterType == typeof(string));

                var method2 = propertyMethod.MakeGenericMethod(typeof(DateTime));
                method2.Invoke(result, new object[] {LastUpdatedAt});
                method2.Invoke(result, new object[] {CreatedAt});
            }


            base.OnModelCreating(m);
        }

        /// <summary>
        /// hook because we need to update the timestamd properties
        /// </summary>
        /// <returns></returns>
        public override int SaveChanges()
        {
            ChangeTracker.DetectChanges();


            //TODO MAYBE do not use reflection (for more performance)
            var updateMethod = this.GetType().GetMethods(BindingFlags.NonPublic | BindingFlags.Instance)
                .First(
                    p =>
                        p.Name == nameof(UpdateTimestampProperties) && p.IsGenericMethod &&
                        p.GetParameters().Length == 0);

            foreach (var type in AllTypes)
            {
                var method = updateMethod.MakeGenericMethod(type);
                method.Invoke(this, new object[] {});
            }
            //UpdateTimestampProperties<ExerciseRelease>();
            //UpdateTimestampProperties<Solution>();
            //UpdateTimestampProperties<SolutionPart>();
            //...

            return base.SaveChanges();
        }

        public override Task<int> SaveChangesAsync(CancellationToken cancellationToken = new CancellationToken())
        {
            ChangeTracker.DetectChanges();
            
            //TODO MAYBE do not use reflection (for more performance)
            var updateMethod = this.GetType().GetMethods(BindingFlags.NonPublic | BindingFlags.Instance)
                .First(
                    p =>
                        p.Name == nameof(UpdateTimestampProperties) && p.IsGenericMethod &&
                        p.GetParameters().Length == 0);

            foreach (var type in AllTypes)
            {
                var method = updateMethod.MakeGenericMethod(type);
                method.Invoke(this, new object[] {});
            }
            //UpdateTimestampProperties<ExerciseRelease>();
            //UpdateTimestampProperties<Solution>();
            //UpdateTimestampProperties<SolutionPart>();
            //...
            
            return base.SaveChangesAsync(cancellationToken);
        }



        /// <summary>
        /// updates the two timestamp properties
        /// </summary>
        /// <typeparam name="T"></typeparam>
        private void UpdateTimestampProperties<T>() where T : class
        {
            var modifiedSourceInfo = ChangeTracker.Entries<T>();

            //times in the db are stored as utc dateTimes
            var now =DateTimeHelper.GetUtcNow();


            foreach (var entry in modifiedSourceInfo)
            {
                if (entry.State == EntityState.Added || entry.State == EntityState.Modified)
                {
                    entry.Property(LastUpdatedAt).CurrentValue = now;

                    if (entry.State == EntityState.Added)
                    {
                        entry.Property(CreatedAt).CurrentValue = now;
                    }
                }
            }
        }
    }
}
