using Microsoft.AspNetCore.Http;

namespace ClientServer.Helpers
{
    public static class NetworkHelper
    {
        /// <summary>
        /// returns the remote ip address
        /// if the X-Forwarded-For or Forwarded header is set then this is used
        /// </summary>
        /// <param name="context"></param>
        /// <returns></returns>
        public static string GetIpAddress(HttpContext context)
        {
            string ip = context.Connection.RemoteIpAddress.ToString();
            
            if (context.Request.Headers.ContainsKey("X-Forwarded-For"))
                ip = context.Request.Headers["X-Forwarded-For"];
            
            //a newer version ... see https://www.nginx.com/resources/wiki/start/topics/examples/forwarded/
            if (context.Request.Headers.ContainsKey("Forwarded"))
                ip = context.Request.Headers["Forwarded"];

            return ip;
        }
    }
}
