﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using BCrypt.Net;
using System.Security.Cryptography;
using System.Text;

namespace ClientServer.Helpers
{
    public static class PasswordHelper
    {
        /// <summary>
        /// never change this just to add some more chars
        /// </summary>
        private const string StaticSalt = "$52346fcy+!_";

        private static Random random = new Random();

        /// <summary>
        /// returns the hash for the pw
        /// </summary>
        /// <param name="pw">the pw to hash</param>
        /// <returns>the hashed pw (using bcrypt)</returns>
        public static string GetHashV0(string pw)
        {
            return BCrypt.Net.BCrypt.HashString(StaticSalt + pw + AppConfiguration.SystemSalt, 13);
        }
        /// <summary>
        /// pwbase64 must be shorter or equal to 72 characters!
        /// </summary>
        /// <param name="pwBase64"></param>
        /// <returns></returns>
        public static string GetHashV1(string pwBase64)
        {
            //salt after pw if salt is too long
            return BCrypt.Net.BCrypt.HashString(pwBase64 + AppConfiguration.SystemSalt, 13);
        }
        /// <summary>
        /// returns the hash with the same hashing algorithm (SHA 256) as the frontend (e.g. to create a new user)
        /// </summary>
        /// <param name="pw"></param>
        /// <returns>always the same length... (44 chars)</returns>
        public static string HashLikeFrontend(string pw)
        {
            string hash;
            using (SHA256 sha = SHA256.Create())
            {
                //hash = BitConverter.ToString(sha.ComputeHash(Encoding.UTF8.GetBytes(pw))).ToLower().Replace("-", "");
                var bytes = sha.ComputeHash(Encoding.UTF8.GetBytes(pw));
                //hash = BitConverter.ToString(sha.ComputeHash(Encoding.UTF8.GetBytes(pw))).ToLower().Replace("-", "");
                // hash = String.Join(String.Empty, bytes.Select(p => p.ToString("x2"))); //needs to be lower case (because cryptojs also uses lower case for output)!
                //sha256 & base64 ensures always the same length
                hash = Convert.ToBase64String(bytes);
            }

            return hash;
        }
        /// <summary>
        /// returns true if same, else not
        /// </summary>
        /// <param name="submittedPw">the plain pw</param>
        /// <param name="hashedPw">the pw from the db</param>
        /// <returns></returns>
        public static bool VerifyV0(string submittedPw, string hashedPw)
        {
            return BCrypt.Net.BCrypt.Verify(StaticSalt + submittedPw + AppConfiguration.SystemSalt, hashedPw);
        }
        
        public static bool VerifyV1(string submittedPwBase64, string hashedPw)
        {
            return BCrypt.Net.BCrypt.Verify(submittedPwBase64 + AppConfiguration.SystemSalt, hashedPw);
        }

        /// <summary>
        /// returns a randomly generated password
        ///
        /// make sure to call <see cref="HashLikeFrontend"/> and <see cref="GetHashV1"/>
        /// 
        /// </summary>
        /// <returns>the generated password</returns>
        public static string GetRandomPassword()
        {
            //this can produce same pws... but the user should change the pw soon anyway 
            return Guid.NewGuid().ToString().Substring(0, 4) + random.Next(10, 99).ToString();

            //this is a bit too long...
//            return Guid.NewGuid().ToString();
        }

    }
}
