﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Models.Exercises.Solution;

namespace ClientServer.Models.Exercises
{
    /// <summary>
    /// a code template file for a <see cref="CodeTemplate"/>
    /// </summary>
    public class TemplateFile
    {
        public int Id { get; set; }
        /// <summary>
        /// the name of the file with extension
        /// </summary>
        [MaxLength(YapexDbContext.DefaultMaxStringLength)]
        public string FileNameWithExtension { get; set; }

        /// <summary>
        /// true: the user can edit this on the client side (and the file name),
        /// false: not
        ///
        /// might be overwritten by IsContentVisibleForUser
        /// </summary>
        public bool IsEditableByUser { get; set; }

        /// <summary>
        /// true: content is visible for the user,
        /// false: the content is set to the empty string before sending to frontend (this file content is not changed in the db),
        ///     normally the initial solution will contain the template file code when the user solution was created
        ///         this is that we have a "working" old user solution even if the exercise changes...
        ///     this only applies for the do exercise view (exercise creators and tutors need to see the content)
        ///     this means that the user cannot edit this file (content or name)
        ///
        /// IF IsEditableByUser is true and this is false the content is hidden and the file is readonly!
        /// </summary>
        public bool IsContentVisibleForUser { get; set; }

        /// <summary>
        /// the order/sorting index how the file is displayed in the editor
        /// -1 not displayed
        /// NOT USED YET
        /// </summary>
        public int DisplayIndex { get; set; }

        /// <summary>
        /// the content of the code template
        /// </summary>
        [MaxLength(YapexDbContext.FileContentMaxStringLength)]
        public string Content { get; set; }


        /// <summary>
        /// the connected <see cref="SolutionFile"/>s, a solution part could be created from a template file
        /// (templates stores whether a file is editable by user or not)
        /// </summary>
        public List<SolutionFile> SolutionParts { get; set; }


    }
}
