using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.Exercises;

namespace ClientServer.Models.Files
{
    /// <summary>
    /// a reference to a file on the hard drive
    /// for exercise description (markdown) asset files
    /// files are stored without an extension because we don't need to know
    ///
    /// as file name we use the id in the db
    /// </summary>
    public class FileReferenceMarkdownAsset : IFileReference
    {
        public int Id { get; set; }

        public FileReferenceMarkdownAsset()
        {
            AssetReferences = new List<ExerciseDescriptionWithFileAsAssetReference>();
        }

        /// <summary>
        /// md5 hash of the file content
        /// </summary>
        [MaxLength(YapexDbContext.DefaultMaxStringLength)]
        public string Hash { get; set; }

        /// <summary>
        /// original name of the file (at the time the file was uploaded)  (might have no extension if uploaded without!)
        /// </summary>
        [MaxLength(YapexDbContext.DefaultMaxStringLength)]
        public string OriginalName { get; set; }
        
        /// <summary>
        /// the mime type (maybe we need to know what content is stored)
        /// </summary>
        [MaxLength(YapexDbContext.DefaultMaxStringLength)]
        public string MimeType { get; set; }

        /// <summary>
        /// the size in bytes of the content (only set once)
        /// </summary>
        public long SizeInBytes { get; set; }

        public DateTime CreatedAt { get; set; }

        public DateTime LastUpdatedAt { get; set; }


        public List<ExerciseDescriptionWithFileAsAssetReference> AssetReferences { get; set; }

    }
}
