﻿using System;
using System.Collections.Generic;
using System.Linq;
using ClientServer.Models.CustomProjects;
using ClientServer.Models.Exercises.Tests;

namespace ClientServer.Models.Interfaces
{
    /// <summary>
    /// a common class for tests and custom tests
    /// </summary>
    public class AllTest
    {

        public AllTest()
        {
            TestAssetsWithContent = new List<AllTestAsset>();
        }
        public int Id { get; set; }
        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the content for the see the test protocol
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// all related <see cref="ITestAsset"/> s e.g. a file to read
        /// </summary>
        public List<AllTestAsset> TestAssetsWithContent { get; set; }

        /// <summary>
        /// the connected <see cref="TestType"/>
        /// </summary>
        public TestType TestType { get; set; }

        public int MaxPoints { get; set; }

        /// <summary>
        /// the connected <see cref="ITestSettings"/>
        /// </summary>
        public ITestSettings TestSettings { get; set; }


        /// <summary>
        /// creates a new all test from teh given test (references are created)
        /// THROWS if a file/asset content could not be found
        /// </summary>
        /// <param name="test">the test</param>
        /// <param name="fileContentDict">the dict with the file content, assetId (reference id) is the key</param>
        /// <returns>the created all test</returns>
        public static AllTest CreateFrom(Test test, Dictionary<int, byte[]> fileContentDict)
        {

            var testAssets = new List<AllTestAsset>();
            
            foreach (var testAssetReference in test.AssetReferences)
            {
                var content = fileContentDict[testAssetReference.FileReferenceTestAssetId];
                if (content == null)
                {
                    //could not be read
                    throw new Exception($"test file could not be read, id: {testAssetReference.FileReferenceTestAssetId}");
                }
                
                testAssets.Add(new AllTestAsset(testAssetReference.FileReferenceTestAssetId, 
                    testAssetReference.FileReferenceTestAsset.OriginalName,
                    content,
                    testAssetReference.FileReferenceTestAsset.MimeType,
                    testAssetReference.FileReferenceTestAsset.Hash
                ));
            }
            
            var allTest = new AllTest()
            {
                Id = test.Id,
                Content = test.Content,
                TestSettings = test.TestSettings,
                DisplayName = test.DisplayName,
                TestAssetsWithContent = testAssets,
                TestType = test.TestType,
                MaxPoints = test.MaxPoints,
            };
            return allTest;
        }
        
        /// <summary>
        /// creates a new all test from teh given test (references are created)
        /// </summary>
        /// <param name="test">the test</param>
        /// <returns>the created all test</returns>
        public static AllTest CreateFrom(CustomProjectTest test, Dictionary<int, byte[]> fileContentDict)
        {
            var testAssets = new List<AllTestAsset>();
            
            foreach (var testAssetReference in test.AssetReferences)
            {
                var content = fileContentDict[testAssetReference.FileReferenceUserFileAssetId];
                if (content == null)
                {
                    //could not be read
                    throw new Exception($"test file could not be read, id: {testAssetReference.FileReferenceUserFileAssetId}");
                }
                
                testAssets.Add(new AllTestAsset(testAssetReference.CustomProjectTestId, 
                    testAssetReference.FileReferenceUserFileAsset.OriginalName,
                    content,
                    testAssetReference.FileReferenceUserFileAsset.MimeType,
                    testAssetReference.FileReferenceUserFileAsset.Hash
                ));
            }
            
            var allTest = new AllTest()
            {
                Id = test.Id,
                Content = test.Content,
                TestSettings = test.CustomProjectTestSettings,
                DisplayName = test.DisplayName,
                TestAssetsWithContent = testAssets,
                TestType = test.TestType,
                MaxPoints = test.MaxPoints,
            };
            return allTest;
        }

        /// <summary>
        /// creates a new all test from teh given test (references are created)
        /// </summary>
        /// <param name="test">the custom test</param>
        /// <param name="testSettings">the default custom test settings</param>
        /// <param name="fileContentDict"></param>
        /// <returns>the created all test</returns>
        public static AllTest CreateFrom(CustomTest test, ITestSettings testSettings, Dictionary<int, byte[]> fileContentDict)
        {
            var testAssets = new List<AllTestAsset>();
            
            foreach (var testAssetReference in test.AssetReferences)
            {
                var content = fileContentDict[testAssetReference.FileReferenceUserFileAssetId];
                if (content == null)
                {
                    //could not be read
                    throw new Exception($"test file could not be read, id: {testAssetReference.FileReferenceUserFileAssetId}");
                }
                
                testAssets.Add(new AllTestAsset(testAssetReference.CustomTestId, 
                    testAssetReference.FileReferenceUserFileAsset.OriginalName,
                    content,
                    testAssetReference.FileReferenceUserFileAsset.MimeType,
                    testAssetReference.FileReferenceUserFileAsset.Hash
                ));
            }
            
            var allTest = new AllTest()
            {
                Id = test.Id,
                Content = test.Content,
                TestSettings = testSettings,
                DisplayName = test.DisplayName,
                TestAssetsWithContent = testAssets,
                TestType = test.TestType,
            };
            return allTest;
        }
    }

    public class AllTestAsset : ITestAsset
    {
        public AllTestAsset(int id, string displayName, byte[] content, string mimeType, string hash)
        {
            Id = id;
            DisplayName = displayName;
            Content = content;
            MimeType = mimeType;
            Hash = hash;
        }

        public int Id { get; set; }
        public string DisplayName { get; set; }
        public byte[] Content { get; set; }
        public string MimeType { get; set; }
        public string Hash { get; set; }
    }
}
